/* Program: ex4-4.c
 * ---------------------------------------------------------------------
 * Solution to exercise 4, series 4.
 *
 * This program checks whether a directed graph is acyclic.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>


/* Definition of a graph with a maximum of N vertices.
   Using an adjacency matrix representation:
   g[u][v] != 0 iff there is an edge (u, v).
 */

#define N 20

typedef unsigned char graph[N][N];


/* Check whether a directed graph is acyclic.

   (This is really hard to check in a directed graph.
   See D. C. Kozen, "The Design and Analysis of Algorithms",
   Springer Verlag, 1991, chapter 4, page 22.)

   We make it a little harder because we want to print the
   cycle, when we find it.  This is more than the exercise
   requires...
   
      n       : the number of nodes in the graph
      g       : the graph's adjacency matrix
      u       : the node that we are visiting
      count   : a pointer to the next visited node's number
      visited : an array where we store the order we visit nodes
      marked  : an array where we store if we've finished with a node
      ccount  : a pointer to the number of nodes in the cycle we found
      cycle   : an array of nodes forming a cycle
      
 */

int hasCycle (unsigned char n, graph g,
              unsigned char u, unsigned char * count,
              unsigned char visited[], unsigned char marked[],
              unsigned char * ccount, unsigned char cycle[])
{
   unsigned char v;

   /* Visit node u */

   visited[u] = (*count)++;

   /* For all outgoing edges (u, v) */

   for (v=0; v<n; v++)
      if (g[u][v]) {

         /* If v has not been visited, check for cycles there first */

         if (!visited[v]) {
            unsigned char w = hasCycle(n, g, v, count, visited, marked,
                                       ccount, cycle);

            /* If a cycle has been found ending with a node visited at
               time w, v is only part of the cycle if it has been visited
               no sooner than w */
            
            if (w) {
               if (visited[v] >= w)
                  cycle[(*ccount)++] = v;
               return w;
            }
         }

         /* If v has been visited before u and has not been marked as
            finished, then we have a cycle ending with the edge (u, v).
            The last node completing the cycle must be v.
          */

         else if (visited[v] <= visited[u] && !marked[v]) {
            cycle[0] = v;
            *ccount = 1;
            return visited[v];
         }

         /* Mark v as finished */

         marked[v] = 1;
      }

   /* If we get so far, there is no cycle */

   return 0;
}


int isAcyclic (unsigned char n, graph g,
               unsigned char * ccount, unsigned char cycle[])
{
   unsigned char visited[N];
   unsigned char marked[N];
   unsigned char count = 1;
   unsigned char u;

   /* Initially, all nodes are not visited and not marked */

   for (u=0; u<n; u++)
      visited[u] = marked[u] = 0;

   /* Starting from the first node, check if there is a cycle at
      the connected subgraph containing u.  Continue until all nodes
      have been visited. */

   for (u=0; u<n; u++)
      if (!visited[u]) {
         unsigned char w = hasCycle(n, g, u, &count, visited, marked,
                                    ccount, cycle);

         /* If a cycle has been found ending with a node visited at
            time w, u is only part of the cycle if it has been visited
            no sooner than w */

         if (w) {
            if (visited[u] >= w)
               cycle[(*ccount)++] = u;
            return 0;
         }
      }

   /* If no cycle has been found, the graph is acyclic */

   return 1;
}


/* Main program */

int main ()
{
   graph g = { /* A  B  C  D  E  F  G  H  I  J */
       /* A */  { 0, 1, 0, 0, 0, 0, 0, 1, 0, 0 },
       /* B */  { 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
       /* C */  { 1, 1, 0, 0, 0, 1, 0, 0, 0, 0 },
       /* D */  { 0, 0, 1, 0, 1, 0, 0, 0, 0, 0 },
       /* E */  { 0, 0, 1, 0, 0, 1, 0, 0, 0, 0 },
       /* F */  { 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
       /* G */  { 0, 0, 0, 0, 0, 0, 0, 1, 0, 0 },
       /* H */  { 0, 0, 0, 0, 0, 0, 0, 0, 1, 0 },
       /* I */  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
       /* J */  { 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 }
   };
   unsigned char cycle[N];
   unsigned char ccount;
   int i;

   if (isAcyclic(10, g, &ccount, cycle))
      printf("The graph is acyclic\n");
   else {
      printf("The graph has a cycle: ");
      for (i=ccount-1; i>=0; i--)
         printf("%c", 'A' + cycle[i]);
      printf("\n");
   }

   return 0;
}
